#!/usr/bin/env python3

import falcon
import os
import sys
import logging
import logging.handlers
import json
import sqlalchemy as sa
import patatt
import smtplib
import email
import email.header
import email.policy
import email.quoprimime
import re
import ezpi
import copy
import textwrap

from configparser import ConfigParser, ExtendedInterpolation
from string import Template
from email import utils
from typing import Tuple, Union, List

from email import charset
charset.add_charset('utf-8', None)
emlpolicy = email.policy.EmailPolicy(utf8=True, cte_type='8bit', max_line_length=None)

qspecials = re.compile(r'[()<>@,:;.\"\[\]]')

DB_VERSION = 1

logger = logging.getLogger('b4-send-receive')
logger.setLevel(logging.DEBUG)


class SendReceiveListener(object):

    def __init__(self, _engine, _config) -> None:
        self._engine = _engine
        self._config = _config
        # You shouldn't use this in production
        if self._engine.driver == 'pysqlite':
            self._init_sa_db()
        logfile = _config['main'].get('logfile')
        loglevel = _config['main'].get('loglevel', 'info')
        if logfile:
            self._init_logger(logfile, loglevel)

    def _init_logger(self, logfile: str, loglevel: str) -> None:
        global logger
        lch = logging.handlers.WatchedFileHandler(os.path.expanduser(logfile))
        lfmt = logging.Formatter('[%(process)d] %(asctime)s - %(levelname)s - %(message)s')
        lch.setFormatter(lfmt)
        if loglevel == 'critical':
            lch.setLevel(logging.CRITICAL)
        elif loglevel == 'debug':
            lch.setLevel(logging.DEBUG)
        else:
            lch.setLevel(logging.INFO)
        logger.addHandler(lch)

    def _init_sa_db(self) -> None:
        logger.info('Setting up SQLite database')
        conn = self._engine.connect()
        md = sa.MetaData()
        meta = sa.Table('meta', md,
                        sa.Column('version', sa.Integer())
                        )
        auth = sa.Table('auth', md,
                        sa.Column('auth_id', sa.Integer(), primary_key=True),
                        sa.Column('created', sa.DateTime(), nullable=False, server_default=sa.sql.func.now()),
                        sa.Column('identity', sa.Text(), nullable=False),
                        sa.Column('selector', sa.Text(), nullable=False),
                        sa.Column('pubkey', sa.Text(), nullable=False),
                        sa.Column('challenge', sa.Text(), nullable=True),
                        sa.Column('verified', sa.Integer(), nullable=False),
                        )
        sa.Index('idx_identity_selector', auth.c.identity, auth.c.selector, unique=True)
        md.create_all(self._engine)
        q = sa.insert(meta).values(version=DB_VERSION)
        conn.execute(q)
        conn.close()

    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        resp.content_type = falcon.MEDIA_TEXT
        resp.text = "We don't serve GETs here\n"

    def send_error(self, resp, message: str) -> None:
        resp.status = falcon.HTTP_500
        logger.critical('Returning error: %s', message)
        resp.text = json.dumps({'result': 'error', 'message': message})

    def send_success(self, resp, message: str) -> None:
        resp.status = falcon.HTTP_200
        logger.debug('Returning success: %s', message)
        resp.text = json.dumps({'result': 'success', 'message': message})

    def get_smtp(self) -> Tuple[Union[smtplib.SMTP, smtplib.SMTP_SSL, None], Tuple[str, str]]:
        sconfig = self._config['sendemail']
        server = sconfig.get('smtpserver', 'localhost')
        port = sconfig.get('smtpserverport', 0)
        encryption = sconfig.get('smtpencryption')

        logger.debug('Connecting to %s:%s', server, port)
        # We only authenticate if we have encryption
        if encryption:
            if encryption in ('tls', 'starttls'):
                # We do startssl
                smtp = smtplib.SMTP(server, port)
                # Introduce ourselves
                smtp.ehlo()
                # Start encryption
                smtp.starttls()
                # Introduce ourselves again to get new criteria
                smtp.ehlo()
            elif encryption in ('ssl', 'smtps'):
                # We do TLS from the get-go
                smtp = smtplib.SMTP_SSL(server, port)
            else:
                raise smtplib.SMTPException('Unclear what to do with smtpencryption=%s' % encryption)

            # If we got to this point, we should do authentication.
            auser = sconfig.get('smtpuser')
            apass = sconfig.get('smtppass')
            if auser and apass:
                # Let any exceptions bubble up
                smtp.login(auser, apass)
        else:
            # We assume you know what you're doing if you don't need encryption
            smtp = smtplib.SMTP(server, port)

        frompair = utils.getaddresses([sconfig.get('from')])[0]
        return smtp, frompair

    def auth_new(self, jdata, resp) -> None:
        # Is it already authorized?
        conn = self._engine.connect()
        md = sa.MetaData()
        identity = jdata.get('identity')
        selector = jdata.get('selector')
        logger.info('New authentication request for %s/%s', identity, selector)
        pubkey = jdata.get('pubkey')
        t_auth = sa.Table('auth', md, autoload=True, autoload_with=self._engine)
        q = sa.select([t_auth.c.auth_id]).where(t_auth.c.identity == identity, t_auth.c.selector == selector,
                                                t_auth.c.verified == 1)
        rp = conn.execute(q)
        if len(rp.fetchall()):
            self.send_error(resp, message='i=%s;s=%s is already authorized' % (identity, selector))
            return
        # delete any existing challenges for this and create a new one
        q = sa.delete(t_auth).where(t_auth.c.identity == identity, t_auth.c.selector == selector,
                                    t_auth.c.verified == 0)
        conn.execute(q)
        # create new challenge
        import uuid
        cstr = str(uuid.uuid4())
        q = sa.insert(t_auth).values(identity=identity, selector=selector, pubkey=pubkey, challenge=cstr,
                                     verified=0)
        conn.execute(q)
        logger.info('Created new challenge for %s/%s: %s', identity, selector, cstr)
        conn.close()
        smtp, frompair = self.get_smtp()
        cmsg = email.message.EmailMessage()
        fromname, fromaddr = frompair
        if len(fromname):
            cmsg.add_header('From', f'{fromname} <{fromaddr}>')
        else:
            cmsg.add_header('From', fromaddr)
        tpt_subject = self._config['templates']['verify-subject'].strip()
        tpt_body = self._config['templates']['verify-body'].strip()
        signature = self._config['templates']['signature'].strip()
        subject = Template(tpt_subject).safe_substitute({'identity': jdata.get('identity')})
        cmsg.add_header('Subject', subject)
        name = jdata.get('name', 'Anonymous Llama')
        cmsg.add_header('To', f'{name} <{identity}>')
        cmsg.add_header('Message-Id', utils.make_msgid('b4-verify'))
        vals = {
            'name': name,
            'myurl': self._config['main'].get('myurl'),
            'challenge': cstr,
        }
        body = Template(tpt_body).safe_substitute(vals)
        body += '\n-- \n'
        body += Template(signature).safe_substitute(vals)
        body += '\n'
        cmsg.set_payload(body, charset='utf-8')
        cmsg.set_charset('utf-8')
        bdata = self.get_msg_as_bytes(cmsg, headers='encode')
        destaddrs = [identity]
        alwaysbcc = self._config['main'].get('alwayscc')
        if alwaysbcc:
            destaddrs += [x[1] for x in utils.getaddresses(alwaysbcc)]
        logger.info('Sending challenge to %s', identity)
        smtp.sendmail(fromaddr, [identity], bdata)
        smtp.close()
        self.send_success(resp, message=f'Challenge generated and sent to {identity}')

    def validate_message(self, conn, t_auth, bdata, verified=1) -> Tuple[str, str, int]:
        # Returns auth_id of the matching record
        pm = patatt.PatattMessage(bdata)
        if not pm.signed:
            raise patatt.ValidationError('Message is not signed')

        auth_id = identity = selector = pubkey = None
        for ds in pm.get_sigs():
            selector = 'default'
            identity = ''
            i = ds.get_field('i')
            if i:
                identity = i.decode()
            s = ds.get_field('s')
            if s:
                selector = s.decode()
            logger.debug('i=%s; s=%s', identity, selector)
            q = sa.select([t_auth.c.auth_id, t_auth.c.pubkey]).where(t_auth.c.identity == identity,
                                                                     t_auth.c.selector == selector,
                                                                     t_auth.c.verified == verified)
            rp = conn.execute(q)
            res = rp.fetchall()
            if res:
                auth_id, pubkey = res[0]
                break

        if not auth_id:
            logger.debug('Did not find a matching identity!')
            raise patatt.NoKeyError('No match for this identity')

        logger.debug('Found matching %s/%s with auth_id=%s', identity, selector, auth_id)
        pm.validate(identity, pubkey.encode())

        return identity, selector, auth_id

    def auth_verify(self, jdata, resp) -> None:
        msg = jdata.get('msg')
        if msg.find('\nverify:') < 0:
            self.send_error(resp, message='Invalid verification message')
            return
        conn = self._engine.connect()
        md = sa.MetaData()
        t_auth = sa.Table('auth', md, autoload=True, autoload_with=self._engine)
        bdata = msg.encode()
        try:
            identity, selector, auth_id = self.validate_message(conn, t_auth, bdata, verified=0)
        except Exception as ex:
            self.send_error(resp, message='Signature validation failed: %s' % ex)
            return
        logger.debug('Message validation passed for %s/%s with auth_id=%s', identity, selector, auth_id)

        # Now compare the challenge to what we received
        q = sa.select([t_auth.c.challenge]).where(t_auth.c.auth_id == auth_id)
        rp = conn.execute(q)
        res = rp.fetchall()
        challenge = res[0][0]
        if msg.find(f'\nverify:{challenge}') < 0:
            self.send_error(resp, message='Challenge verification for %s/%s did not match' % (identity, selector))
            return
        logger.info('Successfully verified challenge for %s/%s with auth_id=%s', identity, selector, auth_id)
        q = sa.update(t_auth).where(t_auth.c.auth_id == auth_id).values(challenge=None, verified=1)
        conn.execute(q)
        conn.close()
        self.send_success(resp, message='Challenge verified for %s/%s' % (identity, selector))

    def auth_delete(self, jdata, resp) -> None:
        msg = jdata.get('msg')
        if msg.find('\nauth-delete') < 0:
            self.send_error(resp, message='Invalid key delete message')
            return
        conn = self._engine.connect()
        md = sa.MetaData()
        t_auth = sa.Table('auth', md, autoload=True, autoload_with=self._engine)
        bdata = msg.encode()
        try:
            identity, selector, auth_id = self.validate_message(conn, t_auth, bdata)
        except Exception as ex:
            self.send_error(resp, message='Signature validation failed: %s' % ex)
            return

        logger.info('Deleting record for %s/%s with auth_id=%s', identity, selector, auth_id)
        q = sa.delete(t_auth).where(t_auth.c.auth_id == auth_id)
        conn.execute(q)
        conn.close()
        self.send_success(resp, message='Record deleted for %s/%s' % (identity, selector))

    def clean_header(self, hdrval: str) -> str:
        if hdrval is None:
            return ''

        decoded = ''
        for hstr, hcs in email.header.decode_header(hdrval):
            if hcs is None:
                hcs = 'utf-8'
            try:
                decoded += hstr.decode(hcs, errors='replace')
            except LookupError:
                # Try as utf-u
                decoded += hstr.decode('utf-8', errors='replace')
            except (UnicodeDecodeError, AttributeError):
                decoded += hstr
        new_hdrval = re.sub(r'\n?\s+', ' ', decoded)
        return new_hdrval.strip()

    def format_addrs(self, pairs: List[Tuple[str, str]], clean: bool = True) -> str:
        addrs = list()
        for pair in pairs:
            if pair[0] == pair[1]:
                addrs.append(pair[1])
                continue
            if clean:
                # Remove any quoted-printable header junk from the name
                pair = (self.clean_header(pair[0]), pair[1])
            # Work around https://github.com/python/cpython/issues/100900
            if not pair[0].startswith('=?') and not pair[0].startswith('"') and qspecials.search(pair[0]):
                quoted = email.utils.quote(pair[0])
                addrs.append(f'"{quoted}" <{pair[1]}>')
                continue
            addrs.append(email.utils.formataddr(pair))
        return ', '.join(addrs)

    def isascii(self, strval: str) -> bool:
        try:
            return strval.isascii()
        except AttributeError:
            return all([ord(c) < 128 for c in strval])

    def wrap_header(self, hdr, width: int = 75, nl: str = '\r\n', transform: str = 'preserve') -> bytes:
        hname, hval = hdr
        if hname.lower() in ('to', 'cc', 'from', 'x-original-from'):
            _parts = [f'{hname}: ']
            first = True
            for addr in email.utils.getaddresses([hval]):
                if transform == 'encode' and not self.isascii(addr[0]):
                    addr = (email.quoprimime.header_encode(addr[0].encode(), charset='utf-8'), addr[1])
                    qp = self.format_addrs([addr], clean=False)
                elif transform == 'decode':
                    qp = self.format_addrs([addr], clean=True)
                else:
                    qp = self.format_addrs([addr], clean=False)
                # See if there is enough room on the existing line
                if first:
                    _parts[-1] += qp
                    first = False
                    continue
                if len(_parts[-1] + ', ' + qp) > width:
                    _parts[-1] += ', '
                    _parts.append(qp)
                    continue
                _parts[-1] += ', ' + qp
        else:
            if transform == 'decode' and hval.find('?=') >= 0:
                hdata = f'{hname}: ' + self.clean_header(hval)
            else:
                hdata = f'{hname}: {hval}'
            if transform != 'encode' or self.isascii(hval):
                if len(hdata) <= width:
                    return hdata.encode()
                # Use simple textwrap, with a small trick that ensures that long non-breakable
                # strings don't show up on the next line from the bare header
                hdata = hdata.replace(': ', ':_', 1)
                wrapped = textwrap.wrap(hdata, break_long_words=False, break_on_hyphens=False,
                                        subsequent_indent=' ', width=width)
                return nl.join(wrapped).replace(':_', ': ', 1).encode()

            qp = f'{hname}: ' + email.quoprimime.header_encode(hval.encode(), charset='utf-8')
            # is it longer than width?
            if len(qp) <= width:
                return qp.encode()

            _parts = list()
            while len(qp) > width:
                wrapat = width - 2
                if len(_parts):
                    # Also allow for the ' ' at the front on continuation lines
                    wrapat -= 1
                # Make sure we don't break on a =XX escape sequence
                while '=' in qp[wrapat - 2:wrapat]:
                    wrapat -= 1
                _parts.append(qp[:wrapat] + '?=')
                qp = ('=?utf-8?q?' + qp[wrapat:])
            _parts.append(qp)
        return f'{nl} '.join(_parts).encode()

    def get_msg_as_bytes(self, msg: email.message.Message, nl: str = '\r\n', headers: str = 'preserve') -> bytes:
        bdata = b''
        for hname, hval in msg.items():
            bdata += self.wrap_header((hname, str(hval)), nl=nl, transform=headers) + nl.encode()
        bdata += nl.encode()
        payload = msg.get_payload(decode=True)
        for bline in payload.split(b'\n'):
            bdata += re.sub(rb'[\r\n]*$', b'', bline) + nl.encode()
        return bdata

    def receive(self, jdata, resp, reflect: bool = False) -> None:
        servicename = self._config['main'].get('myname')
        if not servicename:
            servicename = 'Web Endpoint'
        umsgs = jdata.get('messages')
        if not umsgs:
            self.send_error(resp, message='Missing the messages array')
            return
        logger.debug('Received a request for %s messages', len(umsgs))

        diffre = re.compile(rb'^(---.*\n\+\+\+|GIT binary patch|diff --git \w/\S+ \w/\S+)', flags=re.M | re.I)
        diffstatre = re.compile(rb'^\s*\d+ file.*\d+ (insertion|deletion)', flags=re.M | re.I)

        msgs = list()
        conn = self._engine.connect()
        md = sa.MetaData()
        t_auth = sa.Table('auth', md, autoload=True, autoload_with=self._engine)
        mustdest = self._config['main'].get('mustdest')
        # First, validate all messages
        seenid = identity = selector = validfrom = None
        for umsg in umsgs:
            bdata = umsg.encode()
            try:
                identity, selector, auth_id = self.validate_message(conn, t_auth, bdata)
            except patatt.NoKeyError:
                self.send_error(resp, message='No matching key, please complete web auth first.')
                return
            except Exception as ex:
                self.send_error(resp, message='Signature validation failed: %s' % ex)
                return

            # Make sure only a single auth_id is used within a receive session
            if seenid is None:
                seenid = auth_id
            elif seenid != auth_id:
                self.send_error(resp, message='We only support a single signing identity across patch series.')
                return

            msg = email.message_from_bytes(bdata, policy=emlpolicy)
            logger.debug('Checking sanity on message: %s', msg.get('Subject'))
            # Some quick sanity checking:
            # - Subject must include [PATCH and have only bracketed prefixes before
            # - Content-type may ONLY be text/plain
            # - Has to include a diff or a diffstat
            passes = True
            subject = self.clean_header(msg.get('Subject', ''))
            if not re.match(r'(\[[^]]+\])*\[PATCH', subject):
                passes = False
            if passes:
                cte = msg.get_content_type()
                if cte.lower() != 'text/plain':
                    passes = False
            if passes:
                payload = msg.get_payload(decode=True)
                if not (diffre.search(payload) or diffstatre.search(payload)):
                    passes = False

            if not passes:
                self.send_error(resp, message='This service only accepts patches')
                return

            # Make sure that From, Date, Subject, and Message-Id headers exist
            if not msg.get('From') or not msg.get('Date') or not msg.get('Subject') or not msg.get('Message-Id'):
                self.send_error(resp, message='Message is missing some required headers.')
                return

            # Make sure that From: matches the validated identity. We allow + expansion,
            # such that foo+listname@example.com is allowed for foo@example.com
            allfroms = utils.getaddresses([str(x) for x in msg.get_all('from')])
            # Allow only a single From: address
            if len(allfroms) > 1:
                self.send_error(resp, message='Message may only contain a single From: address.')
                return

            fromaddr = allfroms[0][1]
            if validfrom != fromaddr:
                ldparts = fromaddr.split('@')
                if len(ldparts) != 2:
                    self.send_error(resp, message=f'Invalid address in From: {fromaddr}')
                    return
                lparts = ldparts[0].split('+', maxsplit=1)
                toval = f'{lparts[0]}@{ldparts[1]}'
                if toval != identity:
                    self.send_error(resp, message=f'From header invalid for identity {identity}: {fromaddr}')
                    return
                # usually, all From: addresses will be the same, so use validfrom as a quick bypass
                if validfrom is None:
                    validfrom = fromaddr

            # Check that To/Cc have a mailing list we recognize
            alldests = utils.getaddresses([str(x) for x in msg.get_all('to', [])])
            alldests += utils.getaddresses([str(x) for x in msg.get_all('cc', [])])
            destaddrs = {x[1] for x in alldests}
            if mustdest:
                matched = False
                for destaddr in destaddrs:
                    if re.search(mustdest, destaddr, flags=re.I):
                        matched = True
                        break
                if not matched:
                    self.send_error(resp, message='Destinations must include a mailing list we recognize.')
                    return
            msg.add_header('X-Endpoint-Received', f'by {servicename} for {identity}/{selector} with auth_id={auth_id}')
            msgs.append((msg, destaddrs))

        conn.close()
        # All signatures verified. Prepare messages for sending.
        cfgdomains = self._config['main'].get('mydomains')
        if cfgdomains is not None:
            mydomains = [x.strip() for x in cfgdomains.split(',')]
        else:
            mydomains = list()

        smtp, frompair = self.get_smtp()
        bccaddrs = set()
        _bcc = self._config['main'].get('alwaysbcc')
        if _bcc:
            bccaddrs.update([x[1] for x in utils.getaddresses([_bcc])])

        repo = listid = None
        if 'public-inbox' in self._config and self._config['public-inbox'].get('repo') and not reflect:
            repo = self._config['public-inbox'].get('repo')
            listid = self._config['public-inbox'].get('listid')
            if not os.path.isdir(repo):
                repo = None

        if reflect:
            logger.info('Reflecting %s messages back to %s', len(msgs), identity)
            sentaction = 'Reflected'
        else:
            logger.info('Sending %s messages for %s/%s', len(msgs), identity, selector)
            sentaction = 'Sent'

        for msg, destaddrs in msgs:
            subject = self.clean_header(msg.get('Subject'))
            if repo:
                pmsg = copy.deepcopy(msg)
                if pmsg.get('List-Id'):
                    pmsg.replace_header('List-Id', listid)
                else:
                    pmsg.add_header('List-Id', listid)
                ezpi.add_rfc822(repo, pmsg)
                logger.debug('Wrote %s to public-inbox at %s', subject, repo)

            origfrom = msg.get('From')
            origpair = utils.getaddresses([origfrom])[0]
            origaddr = origpair[1]
            # Does it match one of our domains
            mydomain = False
            for _domain in mydomains:
                if origaddr.endswith(f'@{_domain}'):
                    mydomain = True
                    break
            if mydomain:
                logger.debug('%s matches mydomain, no substitution required', origaddr)
                fromaddr = origaddr
            else:
                logger.debug('%s does not match mydomain, substitution required', origaddr)
                # We can't just send this as-is due to DMARC policies. Therefore, we set
                # Reply-To and X-Original-From.
                fromaddr = frompair[1]
                origname = origpair[0]
                if not origname:
                    origname = origpair[1]
                delim = self._config['main'].get('from-recipient-delimiter', '+')
                if delim and '@' in fromaddr:
                    _flocal, _fdomain = fromaddr.split('@', maxsplit=1)
                    _forig = origaddr.replace('@', '.')
                    fromaddr = f'{_flocal}{delim}{_forig}@{_fdomain}'
                msg.replace_header('From', f'{origname} via {servicename} <{fromaddr}>')

                if msg.get('X-Original-From'):
                    msg.replace_header('X-Original-From', origfrom)
                else:
                    msg.add_header('X-Original-From', origfrom)
                if msg.get('Reply-To'):
                    msg.replace_header('Reply-To', f'<{origpair[1]}>')
                else:
                    msg.add_header('Reply-To', f'<{origpair[1]}>')

                body = msg.get_payload(decode=True)
                # Add a From: header (if there isn't already one), but only if it's a patch
                if diffre.search(body):
                    # Parse it as a message and see if we get a From: header
                    cmsg = email.message_from_bytes(body, policy=emlpolicy)
                    if cmsg.get('From') is None:
                        newbody = 'From: ' + self.clean_header(origfrom) + '\n'
                        if cmsg.get('Subject'):
                            newbody += 'Subject: ' + self.clean_header(cmsg.get('Subject')) + '\n'
                        if cmsg.get('Date'):
                            newbody += 'Date: ' + self.clean_header(cmsg.get('Date')) + '\n'
                        newbody += '\n' + body.decode()
                        msg.set_payload(newbody, charset='utf-8')
                        # If we have non-ascii content in the new body, force CTE to 8bit
                        if msg['Content-Transfer-Encoding'] == '7bit' and not all(ord(char) < 128 for char in newbody):
                            msg.set_charset('utf-8')
                            msg.replace_header('Content-Transfer-Encoding', '8bit')

            if bccaddrs:
                destaddrs.update(bccaddrs)

            if not self._config['main'].getboolean('dryrun'):
                bdata = self.get_msg_as_bytes(msg, headers='encode')
                if reflect:
                    smtp.sendmail(fromaddr, [identity], bdata)
                else:
                    smtp.sendmail(fromaddr, list(destaddrs), bdata)
                logger.info('%s: %s', sentaction, subject)
            else:
                logger.info('---DRYRUN MSG START---')
                logger.info(msg)
                logger.info('---DRYRUN MSG END---')

        smtp.close()
        if repo:
            # run it once after writing all messages
            logger.debug('Running public-inbox repo hook (if present)')
            ezpi.run_hook(repo)
        logger.info('%s %s messages for %s/%s', sentaction, len(msgs), identity, selector)
        self.send_success(resp, message=f'{sentaction} {len(msgs)} messages for {identity}/{selector}')

    def on_post(self, req, resp):
        if not req.content_length:
            resp.status = falcon.HTTP_500
            resp.content_type = falcon.MEDIA_TEXT
            resp.text = 'Payload required\n'
            return
        raw = req.bounded_stream.read()
        try:
            jdata = json.loads(raw)
        except Exception:
            resp.status = falcon.HTTP_500
            resp.content_type = falcon.MEDIA_TEXT
            resp.text = 'Failed to parse the request\n'
            return
        action = jdata.get('action')
        if not action:
            logger.critical('Action not set from %s', req.remote_addr)

        logger.info('Action: %s; from: %s', action, req.remote_addr)
        if action == 'auth-new':
            self.auth_new(jdata, resp)
            return
        if action == 'auth-verify':
            self.auth_verify(jdata, resp)
            return
        if action == 'auth-delete':
            self.auth_delete(jdata, resp)
            return
        if action == 'receive':
            self.receive(jdata, resp)
            return
        if action == 'reflect':
            self.receive(jdata, resp, reflect=True)
            return

        resp.status = falcon.HTTP_500
        resp.content_type = falcon.MEDIA_TEXT
        resp.text = 'Unknown action: %s\n' % action


parser = ConfigParser(interpolation=ExtendedInterpolation())
cfgfile = os.getenv('CONFIG')
if not cfgfile or not os.path.exists(cfgfile):
    sys.stderr.write('CONFIG env var is not set or is not valid')
    sys.exit(1)

parser.read(cfgfile)

gpgbin = parser['main'].get('gpgbin')
if gpgbin:
    patatt.GPGBIN = gpgbin

dburl = parser['main'].get('dburl')
# By default, recycle db connections after 5 min
db_pool_recycle = parser['main'].getint('dbpoolrecycle', 300)
engine = sa.create_engine(dburl, pool_recycle=db_pool_recycle)
srl = SendReceiveListener(engine, parser)
app = falcon.App()
mp = os.getenv('MOUNTPOINT', '/_b4_submit')
app.add_route(mp, srl)


if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(message)s')
    ch.setFormatter(formatter)
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)

    with make_server('', 8000, app) as httpd:
        logger.info('Serving on port 8000...')

        # Serve until process is killed
        httpd.serve_forever()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024 by the Linux Foundation
#
"""Headless Textual tests for review-TUI modal screens.

Uses Textual's built-in ``App.run_test()`` / ``Pilot`` harness so the
tests run without a real terminal.  Only lightweight, self-contained
modals are exercised here — no database, network, or git needed.
"""
import pytest

from typing import Any, Dict, List, Optional, Tuple

from textual.app import App, ComposeResult
from textual.widgets import Input, Label, ListView

from b4.review_tui._modals import (
    ActionScreen,
    ConfirmScreen,
    HelpScreen,
    LimitScreen,
    NoteScreen,
    PriorReviewScreen,
    RevisionChoiceScreen,
    SetStateScreen,
    SnoozeScreen,
    TrailerScreen,
    UpdateRevisionScreen,
    TRACKING_HELP_LINES,
)


# ---------------------------------------------------------------------------
# Compat helper — Textual ≥ 1.0 (pip) uses Static.content,
# older builds (e.g. Fedora 43 package) still use Static.renderable.
# ---------------------------------------------------------------------------

def _static_text(widget: Any) -> str:
    """Return the text content of a Static widget across Textual versions."""
    if hasattr(widget, 'content'):
        return str(widget.content)
    return str(widget.renderable)


# ---------------------------------------------------------------------------
# Minimal host app — just enough to push modal screens onto
# ---------------------------------------------------------------------------

class ModalTestApp(App[None]):
    """Bare app that serves as a host for pushing modal screens."""

    def compose(self) -> ComposeResult:
        yield Label('host')


# ---------------------------------------------------------------------------
# HelpScreen
# ---------------------------------------------------------------------------

class TestHelpScreen:
    """Tests for the HelpScreen modal."""

    @staticmethod
    def _lines() -> List[str]:
        return [
            '[bold]Test Help[/bold]\n',
            '\n',
            '  [bold]a[/bold]  Do alpha\n',
            '  [bold]b[/bold]  Do bravo\n',
        ]

    @pytest.mark.asyncio
    async def test_escape_dismisses(self) -> None:
        app = ModalTestApp()
        dismissed: List[Any] = []

        async with app.run_test() as pilot:
            app.push_screen(HelpScreen(self._lines()), dismissed.append)
            await pilot.pause()
            # The help screen should now be on top
            assert isinstance(app.screen, HelpScreen)

            await pilot.press('escape')
            await pilot.pause()
            # Should be back on the host screen
            assert not isinstance(app.screen, HelpScreen)
            assert dismissed == [None]

    @pytest.mark.asyncio
    async def test_question_mark_dismisses(self) -> None:
        app = ModalTestApp()
        dismissed: List[Any] = []

        async with app.run_test() as pilot:
            app.push_screen(HelpScreen(self._lines()), dismissed.append)
            await pilot.pause()

            await pilot.press('question_mark')
            await pilot.pause()
            assert not isinstance(app.screen, HelpScreen)
            assert dismissed == [None]

    @pytest.mark.asyncio
    async def test_q_dismisses(self) -> None:
        app = ModalTestApp()
        dismissed: List[Any] = []

        async with app.run_test() as pilot:
            app.push_screen(HelpScreen(self._lines()), dismissed.append)
            await pilot.pause()

            await pilot.press('q')
            await pilot.pause()
            assert not isinstance(app.screen, HelpScreen)
            assert dismissed == [None]

    @pytest.mark.asyncio
    async def test_scroll_bindings_do_not_dismiss(self) -> None:
        """j, k, space, etc. should scroll but keep the modal open."""
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(HelpScreen(TRACKING_HELP_LINES))
            await pilot.pause()
            assert isinstance(app.screen, HelpScreen)

            for key in ('j', 'k', 'down', 'up', 'space', 'backspace',
                        'pagedown', 'pageup'):
                await pilot.press(key)
                await pilot.pause()
                assert isinstance(app.screen, HelpScreen), \
                    f'{key!r} unexpectedly closed the help screen'

    @pytest.mark.asyncio
    async def test_content_rendered(self) -> None:
        """The static content inside the modal should contain our text."""
        app = ModalTestApp()

        async with app.run_test() as pilot:
            lines = self._lines()
            app.push_screen(HelpScreen(lines))
            await pilot.pause()

            dialog = app.screen.query_one('#help-dialog')
            assert dialog is not None


# ---------------------------------------------------------------------------
# ConfirmScreen
# ---------------------------------------------------------------------------

class TestConfirmScreen:
    """Tests for the ConfirmScreen modal."""

    @pytest.mark.asyncio
    async def test_y_confirms(self) -> None:
        app = ModalTestApp()
        results: List[Optional[bool]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                ConfirmScreen('Delete?', ['This is permanent.']),
                results.append,
            )
            await pilot.pause()
            assert isinstance(app.screen, ConfirmScreen)

            await pilot.press('y')
            await pilot.pause()
            assert not isinstance(app.screen, ConfirmScreen)
            assert results == [True]

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[bool]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                ConfirmScreen('Delete?', ['This is permanent.']),
                results.append,
            )
            await pilot.pause()

            await pilot.press('escape')
            await pilot.pause()
            assert not isinstance(app.screen, ConfirmScreen)
            assert results == [False]

    @pytest.mark.asyncio
    async def test_body_lines_rendered(self) -> None:
        """Each body string should appear as a Static widget."""
        app = ModalTestApp()
        body = ['Line one.', 'Line two.', 'Line three.']

        async with app.run_test() as pilot:
            app.push_screen(ConfirmScreen('Title', body))
            await pilot.pause()

            statics = app.screen.query('#confirm-dialog Static')
            # body lines + hint line + possibly title
            rendered = [_static_text(s) for s in statics]
            for line in body:
                assert any(line in r for r in rendered), \
                    f'{line!r} not found in rendered statics'

    @pytest.mark.asyncio
    async def test_subject_shown(self) -> None:
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(
                ConfirmScreen('Title', ['body'], subject='My Subject'),
            )
            await pilot.pause()

            title_widget = app.screen.query_one('#confirm-title')
            assert 'My Subject' in _static_text(title_widget)

    @pytest.mark.asyncio
    async def test_warning_border(self) -> None:
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(
                ConfirmScreen('Oops', ['careful'], border='$warning'),
            )
            await pilot.pause()

            dialog = app.screen.query_one('#confirm-dialog')
            assert dialog.has_class('--border-warning')


# ---------------------------------------------------------------------------
# TrailerScreen
# ---------------------------------------------------------------------------

class TestTrailerScreen:
    """Tests for the TrailerScreen modal."""

    @pytest.mark.asyncio
    async def test_cancel_returns_none(self) -> None:
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(TrailerScreen([]), results.append)
            await pilot.pause()
            assert isinstance(app.screen, TrailerScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_confirm_empty_returns_empty_list(self) -> None:
        """With nothing toggled, q/confirm returns an empty list."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(TrailerScreen([]), results.append)
            await pilot.pause()

            await pilot.press('q')
            await pilot.pause()
            assert len(results) == 1
            assert results[0] == []

    @pytest.mark.asyncio
    async def test_existing_trailers_pretoggled(self) -> None:
        """Trailers already present should be pre-selected."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                TrailerScreen(['Reviewed-by: Alice <a@b.com>']),
                results.append,
            )
            await pilot.pause()

            # Confirm immediately — should have Reviewed-by selected
            await pilot.press('q')
            await pilot.pause()
            assert results[0] == ['Reviewed-by']

    @pytest.mark.asyncio
    async def test_toggle_and_confirm(self) -> None:
        """Space toggles the highlighted item; q confirms."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(TrailerScreen([]), results.append)
            await pilot.pause()

            # First item is Acked-by (index 0), toggle it
            await pilot.press('space')
            await pilot.pause()
            # Move down to Reviewed-by, toggle it
            await pilot.press('j')
            await pilot.press('space')
            await pilot.pause()
            # Confirm
            await pilot.press('q')
            await pilot.pause()

            assert len(results) == 1
            assert results[0] is not None
            assert 'Acked-by' in results[0]
            assert 'Reviewed-by' in results[0]
            assert 'Tested-by' not in results[0]
            assert 'NACKed-by' not in results[0]

    @pytest.mark.asyncio
    async def test_toggle_twice_deselects(self) -> None:
        """Toggling the same item twice should deselect it."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(TrailerScreen([]), results.append)
            await pilot.pause()

            await pilot.press('space')  # select Acked-by
            await pilot.press('space')  # deselect Acked-by
            await pilot.pause()

            await pilot.press('q')
            await pilot.pause()
            assert results[0] == []

    @pytest.mark.asyncio
    async def test_jk_navigation(self) -> None:
        """j/k should move the highlight without toggling."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(TrailerScreen([]), results.append)
            await pilot.pause()

            # Navigate down to Tested-by (index 2) and toggle only it
            await pilot.press('j')
            await pilot.press('j')
            await pilot.press('space')
            await pilot.pause()

            await pilot.press('q')
            await pilot.pause()
            assert results[0] == ['Tested-by']

    @pytest.mark.asyncio
    async def test_enter_confirms(self) -> None:
        """Enter on the ListView should also confirm (via on_list_view_selected)."""
        app = ModalTestApp()
        results: List[Optional[List[str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                TrailerScreen(['Acked-by: Bob <b@c.com>']),
                results.append,
            )
            await pilot.pause()

            await pilot.press('enter')
            await pilot.pause()
            assert len(results) == 1
            assert results[0] == ['Acked-by']


# ---------------------------------------------------------------------------
# NoteScreen
# ---------------------------------------------------------------------------

class TestNoteScreen:
    """Tests for the NoteScreen modal."""

    @staticmethod
    def _entries() -> List[Tuple[str, str, str]]:
        return [
            ('Alice <alice@example.com>', 'green', 'Looks good to me.'),
            ('Bob <bob@example.com>', 'blue', 'Needs a rebase.'),
        ]

    @pytest.mark.asyncio
    async def test_escape_returns_none(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(NoteScreen(self._entries()), results.append)
            await pilot.pause()
            assert isinstance(app.screen, NoteScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_e_returns_edit(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(NoteScreen(self._entries()), results.append)
            await pilot.pause()

            await pilot.press('e')
            await pilot.pause()
            assert results == ['__EDIT__']

    @pytest.mark.asyncio
    async def test_d_returns_delete(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(NoteScreen(self._entries()), results.append)
            await pilot.pause()

            await pilot.press('d')
            await pilot.pause()
            assert results == ['__DELETE__']

    @pytest.mark.asyncio
    async def test_empty_notes(self) -> None:
        """Modal should still render with no entries."""
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(NoteScreen([]))
            await pilot.pause()
            assert isinstance(app.screen, NoteScreen)
            assert app.screen.query_one('#note-viewer') is not None


# ---------------------------------------------------------------------------
# PriorReviewScreen
# ---------------------------------------------------------------------------

class TestPriorReviewScreen:
    """Tests for the PriorReviewScreen modal."""

    @pytest.mark.asyncio
    async def test_escape_dismisses(self) -> None:
        app = ModalTestApp()
        results: List[Any] = []

        async with app.run_test() as pilot:
            app.push_screen(
                PriorReviewScreen('== Patch 1 ==\nLGTM'),
                results.append,
            )
            await pilot.pause()
            assert isinstance(app.screen, PriorReviewScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert not isinstance(app.screen, PriorReviewScreen)
            assert results == [None]

    @pytest.mark.asyncio
    async def test_content_rendered(self) -> None:
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(PriorReviewScreen('== Header ==\nSome text'))
            await pilot.pause()
            assert app.screen.query_one('#prior-review-viewer') is not None


# ---------------------------------------------------------------------------
# RevisionChoiceScreen
# ---------------------------------------------------------------------------

class TestRevisionChoiceScreen:
    """Tests for the RevisionChoiceScreen modal."""

    @pytest.mark.asyncio
    async def test_n_returns_newest(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                RevisionChoiceScreen(current_rev=2, newest_rev=5),
                results.append,
            )
            await pilot.pause()
            assert isinstance(app.screen, RevisionChoiceScreen)

            await pilot.press('n')
            await pilot.pause()
            assert results == [5]

    @pytest.mark.asyncio
    async def test_o_returns_current(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                RevisionChoiceScreen(current_rev=2, newest_rev=5),
                results.append,
            )
            await pilot.pause()

            await pilot.press('o')
            await pilot.pause()
            assert results == [2]

    @pytest.mark.asyncio
    async def test_escape_returns_none(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                RevisionChoiceScreen(current_rev=1, newest_rev=3),
                results.append,
            )
            await pilot.pause()

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]


# ---------------------------------------------------------------------------
# SnoozeScreen
# ---------------------------------------------------------------------------

class TestSnoozeScreen:
    """Tests for the SnoozeScreen modal."""

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()
            assert isinstance(app.screen, SnoozeScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_duration_snooze(self) -> None:
        """Entering a duration and confirming should return a datetime."""
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()

            # The duration input should be focused by default
            dur_input = app.screen.query_one('#snooze-duration', Input)
            dur_input.value = '3d'

            await pilot.press('ctrl+y')
            await pilot.pause()

            assert len(results) == 1
            assert results[0] is not None
            assert results[0]['source'] == 'duration'
            assert results[0]['input'] == '3d'
            # until should be an ISO datetime
            assert 'T' in results[0]['until']

    @pytest.mark.asyncio
    async def test_tag_snooze(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()

            tag_input = app.screen.query_one('#snooze-tag', Input)
            tag_input.value = 'v6.15-rc3'

            await pilot.press('ctrl+y')
            await pilot.pause()

            assert len(results) == 1
            assert results[0] is not None
            assert results[0]['until'] == 'tag:v6.15-rc3'
            assert results[0]['source'] == 'tag'

    @pytest.mark.asyncio
    async def test_empty_fields_shows_error(self) -> None:
        """Confirming with no fields filled should show an error, not dismiss."""
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()

            await pilot.press('ctrl+y')
            await pilot.pause()
            # Should still be on the snooze screen (not dismissed)
            assert isinstance(app.screen, SnoozeScreen)
            assert len(results) == 0

            error = app.screen.query_one('#snooze-error')
            error_text = _static_text(error).lower()
            assert 'enter' in error_text or 'please' in error_text

    @pytest.mark.asyncio
    async def test_multiple_fields_shows_error(self) -> None:
        """Filling more than one field should show an error."""
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()

            app.screen.query_one('#snooze-duration', Input).value = '1d'
            app.screen.query_one('#snooze-tag', Input).value = 'v6.15'

            await pilot.press('ctrl+y')
            await pilot.pause()
            assert isinstance(app.screen, SnoozeScreen)
            assert len(results) == 0

            error = app.screen.query_one('#snooze-error')
            assert 'only one' in _static_text(error).lower()

    @pytest.mark.asyncio
    async def test_invalid_duration_shows_error(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Dict[str, str]]] = []

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(), results.append)
            await pilot.pause()

            app.screen.query_one('#snooze-duration', Input).value = 'banana'

            await pilot.press('ctrl+y')
            await pilot.pause()
            assert isinstance(app.screen, SnoozeScreen)
            assert len(results) == 0

    @pytest.mark.asyncio
    async def test_prepopulate_last_source(self) -> None:
        """Re-opening with last_source/last_input should pre-fill the field."""
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(SnoozeScreen(last_source='tag', last_input='v6.14'))
            await pilot.pause()

            tag_input = app.screen.query_one('#snooze-tag', Input)
            assert tag_input.value == 'v6.14'


# ---------------------------------------------------------------------------
# SetStateScreen
# ---------------------------------------------------------------------------

class TestSetStateScreen:
    """Tests for the SetStateScreen modal."""

    @staticmethod
    def _states() -> List[Dict[str, Any]]:
        return [
            {'slug': 'new', 'name': 'New'},
            {'slug': 'reviewing', 'name': 'Reviewing'},
            {'slug': 'replied', 'name': 'Replied'},
            {'slug': 'waiting', 'name': 'Waiting'},
        ]

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Tuple[str, bool]]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                SetStateScreen(self._states(), 'new'),
                results.append,
            )
            await pilot.pause()
            assert isinstance(app.screen, SetStateScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_select_state_with_enter(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Tuple[str, bool]]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                SetStateScreen(self._states(), 'new'),
                results.append,
            )
            await pilot.pause()

            # Navigate to 'reviewing' (one down from 'new')
            await pilot.press('j')
            await pilot.press('enter')
            await pilot.pause()

            assert len(results) == 1
            assert results[0] is not None
            assert results[0][0] == 'reviewing'
            assert results[0][1] is False  # not archived

    @pytest.mark.asyncio
    async def test_archive_toggle(self) -> None:
        app = ModalTestApp()
        results: List[Optional[Tuple[str, bool]]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                SetStateScreen(self._states(), 'new'),
                results.append,
            )
            await pilot.pause()

            # Toggle archive, then confirm
            await pilot.press('a')
            await pilot.press('enter')
            await pilot.pause()

            assert len(results) == 1
            assert results[0] is not None
            assert results[0][1] is True  # archived

    @pytest.mark.asyncio
    async def test_current_state_preselected(self) -> None:
        """The ListView index should start on the current state."""
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(SetStateScreen(self._states(), 'replied'))
            await pilot.pause()

            lv = app.screen.query_one('#state-list', ListView)
            assert lv.index == 2  # 'replied' is at index 2


# ---------------------------------------------------------------------------
# LimitScreen
# ---------------------------------------------------------------------------

class TestLimitScreen:
    """Tests for the LimitScreen modal."""

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(LimitScreen(), results.append)
            await pilot.pause()
            assert isinstance(app.screen, LimitScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_enter_submits_value(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(LimitScreen(), results.append)
            await pilot.pause()

            inp = app.screen.query_one('#limit-input', Input)
            inp.value = 'netfilter'

            await pilot.press('enter')
            await pilot.pause()
            assert results == ['netfilter']

    @pytest.mark.asyncio
    async def test_empty_enter_clears_filter(self) -> None:
        """Submitting empty input should return empty string (clear filter)."""
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(LimitScreen(), results.append)
            await pilot.pause()

            await pilot.press('enter')
            await pilot.pause()
            assert results == ['']

    @pytest.mark.asyncio
    async def test_current_pattern_prepopulated(self) -> None:
        app = ModalTestApp()

        async with app.run_test() as pilot:
            app.push_screen(LimitScreen(current_pattern='drm'))
            await pilot.pause()

            inp = app.screen.query_one('#limit-input', Input)
            assert inp.value == 'drm'


# ---------------------------------------------------------------------------
# ActionScreen
# ---------------------------------------------------------------------------

class TestActionScreen:
    """Tests for the ActionScreen modal."""

    @staticmethod
    def _actions() -> List[Tuple[str, str]]:
        return [
            ('review', 'Review this series'),
            ('take', 'Take (apply) this series'),
            ('snooze', 'Snooze this series'),
            ('archive', 'Archive this series'),
        ]

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()
            assert isinstance(app.screen, ActionScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_enter_confirms_highlighted(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()

            # First item is 'review'
            await pilot.press('enter')
            await pilot.pause()
            assert results == ['review']

    @pytest.mark.asyncio
    async def test_navigate_and_confirm(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()

            await pilot.press('j')  # take
            await pilot.press('j')  # snooze
            await pilot.press('enter')
            await pilot.pause()
            assert results == ['snooze']

    @pytest.mark.asyncio
    async def test_shortcut_key_selects_directly(self) -> None:
        """Pressing the shortcut char should immediately dismiss."""
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()

            # 'T' is the shortcut for 'take'
            await pilot.press('T')
            await pilot.pause()
            assert results == ['take']

    @pytest.mark.asyncio
    async def test_shortcut_r_for_review(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()

            await pilot.press('r')
            await pilot.pause()
            assert results == ['review']

    @pytest.mark.asyncio
    async def test_shortcut_x_for_archive(self) -> None:
        app = ModalTestApp()
        results: List[Optional[str]] = []

        async with app.run_test() as pilot:
            app.push_screen(ActionScreen(self._actions()), results.append)
            await pilot.pause()

            await pilot.press('x')
            await pilot.pause()
            assert results == ['archive']


# ---------------------------------------------------------------------------
# UpdateRevisionScreen
# ---------------------------------------------------------------------------

class TestUpdateRevisionScreen:
    """Tests for the UpdateRevisionScreen modal."""

    @staticmethod
    def _revisions() -> List[Dict[str, Any]]:
        return [
            {'revision': 2, 'subject': '[PATCH v2] fix the thing'},
            {'revision': 3, 'subject': '[PATCH v3] fix the thing properly'},
            {'revision': 4, 'subject': '[PATCH v4] fix the thing for real'},
        ]

    @pytest.mark.asyncio
    async def test_escape_cancels(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                UpdateRevisionScreen(1, self._revisions()),
                results.append,
            )
            await pilot.pause()
            assert isinstance(app.screen, UpdateRevisionScreen)

            await pilot.press('escape')
            await pilot.pause()
            assert results == [None]

    @pytest.mark.asyncio
    async def test_select_first_revision(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                UpdateRevisionScreen(1, self._revisions()),
                results.append,
            )
            await pilot.pause()

            await pilot.press('enter')
            await pilot.pause()
            assert results == [2]

    @pytest.mark.asyncio
    async def test_navigate_and_select(self) -> None:
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            app.push_screen(
                UpdateRevisionScreen(1, self._revisions()),
                results.append,
            )
            await pilot.pause()

            await pilot.press('j')  # v3
            await pilot.press('j')  # v4
            await pilot.press('enter')
            await pilot.pause()
            assert results == [4]

    @pytest.mark.asyncio
    async def test_filters_older_revisions(self) -> None:
        """Only revisions newer than current should appear."""
        app = ModalTestApp()
        results: List[Optional[int]] = []

        async with app.run_test() as pilot:
            # current_revision=3, so only v4 should appear
            app.push_screen(
                UpdateRevisionScreen(3, self._revisions()),
                results.append,
            )
            await pilot.pause()

            # Only one item in the list — entering should select v4
            await pilot.press('enter')
            await pilot.pause()
            assert results == [4]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024 by the Linux Foundation
#
__author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'

import email.message
import email.parser
import email.policy
import email.utils
import json
import os
import subprocess
import tempfile
import unicodedata

from typing import Any, Dict, List, Optional, Set, Tuple

import b4
import b4.mbox
import b4.review
import b4.review.tracking

from collections import defaultdict

from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Footer, ListView, RichLog
from textual.widgets._footer import FooterKey
from rich import box
from rich.padding import Padding
from rich.panel import Panel
from rich.rule import Rule
from rich.text import Text

logger = b4.logger


def get_thread_msgs(
    topdir: str,
    message_id: str,
    blob_sha: str = '',
    quiet: bool = False,
) -> Optional[List[email.message.EmailMessage]]:
    """Retrieve thread messages, trying the local git blob first.

    If *blob_sha* points to a valid mbox blob in the repository, the
    thread is loaded from there without network access.  Otherwise
    falls back to fetching from lore via ``b4.get_pi_thread_by_msgid``.

    Returns a list of ``EmailMessage`` objects, or *None* on failure.
    """
    if blob_sha:
        mbox_bytes = b4.review.tracking.get_thread_mbox(topdir, blob_sha)
        if mbox_bytes:
            msgs = b4.mailsplit_bytes(
                mbox_bytes,
                os.path.join(topdir, '.git', 'b4-thread-tmp'))
            if msgs:
                return msgs

    return b4.get_pi_thread_by_msgid(message_id, quiet=quiet) or None


# Per-patch state indicators — same glyphs as _tracking_app._STATUS_SYMBOLS
PATCH_STATE_MARKERS: Dict[str, str] = {
    '':          ' ',
    'external':  '\u00b1',  # ± plus-minus   (= external comments available)
    'draft':     '\u270e',  # ✎ pencil       (= maintainer reviewing)
    'done':      '\u2713',  # ✓ check        (= done)
    'skip':      '\u2715',  # ✕ cross        (= skipped)
    'unchanged': '\u2261',  # ≡ identical-to (= unchanged from prior rev)
}

# CI check label text (colour-free constant — used with ci_check_styles())
CI_CHECK_LABELS = {
    'pending': '\u25cf pending',
    'success': '\u25cf pass',
    'warning': '\u25cf warning',
    'fail': '\u25cf FAIL',
}


def display_width(s: str) -> int:
    """Return the terminal display width of *s*, accounting for full-width chars."""
    w = 0
    for ch in s:
        w += 2 if unicodedata.east_asian_width(ch) in ('F', 'W') else 1
    return w


def pad_display(s: str, width: int) -> str:
    """Pad or truncate *s* to *width* terminal columns, accounting for full-width chars."""
    dw = display_width(s)
    if dw > width:
        # Truncate with ellipsis
        truncated: List[str] = []
        tw = 0
        for ch in s:
            cw = 2 if unicodedata.east_asian_width(ch) in ('F', 'W') else 1
            if tw + cw > width - 1:
                break
            truncated.append(ch)
            tw += cw
        return ''.join(truncated) + '\u2026' + ' ' * (width - tw - 1)
    if dw < width:
        return s + ' ' * (width - dw)
    return s


def _fix_ansi_theme(app: Any) -> None:
    """Work around Textual theme-watcher bug.

    The ``theme`` reactive in Textual's ``App`` uses ``init=False``,
    so ``_watch_theme()`` never fires for the initial theme set via
    ``TEXTUAL_THEME`` env-var.  This leaves ``ansi_color`` as False
    and all ``:ansi`` CSS pseudo-class overrides dead.  Call this
    from ``on_mount()`` to force the watcher when needed.
    """
    if app.current_theme.name == 'textual-ansi' and not app.ansi_color:
        app._watch_theme(app.theme)


def _to_rich_color(textual_color: str) -> str:
    """Convert a Textual CSS colour value to a Rich-compatible name.

    Textual uses ``ansi_green``, ``ansi_bright_blue``, etc. in its CSS
    variable system.  Rich expects ``green``, ``bright_blue``, etc.
    Non-ansi values (hex codes, named CSS colours) pass through unchanged.
    ``ansi_default`` maps to ``default``.
    """
    if textual_color.startswith('ansi_'):
        return textual_color[5:]  # strip 'ansi_' prefix
    return textual_color


def resolve_styles(app: Any) -> Dict[str, str]:
    """Resolve Textual CSS variables into Rich-compatible colour strings.

    Call this once per render cycle and pass the resulting dict to
    helper functions like ``_write_diff_line()`` and ``ci_styles()``.

    The dict maps semantic names to colour strings that Rich ``Text``
    objects can use directly in *style* parameters.
    """
    v = app.get_css_variables()
    return {
        'success': _to_rich_color(v.get('success', 'green')),
        'error': _to_rich_color(v.get('error', 'red')),
        'warning': _to_rich_color(v.get('warning', 'dark_orange')),
        'accent': _to_rich_color(v.get('accent', 'cyan')),
        'secondary': _to_rich_color(v.get('secondary-lighten-3', 'magenta')),
        'foreground': _to_rich_color(v.get('foreground', 'bright_white')),
        'panel': _to_rich_color(v.get('panel', 'grey11')),
        'surface': _to_rich_color(v.get('surface', '#1e1e1e')),
        'primary': _to_rich_color(v.get('primary', 'dark_blue')),
        'text-muted': _to_rich_color(v.get('text-muted', 'grey70')),
        'syntax_theme': 'ansi_dark' if app.current_theme.dark else 'ansi_light',
    }


def ci_styles(ts: Dict[str, str]) -> Dict[str, str]:
    """Return CI indicator styles from a resolved theme dict."""
    return {
        'pending': 'dim',
        'success': ts['success'],
        'warning': ts['warning'],
        'fail': f"bold {ts['error']}",
    }


def ci_markup(ts: Dict[str, str]) -> Dict[str, str]:
    """Return CI dot markup strings from a resolved theme dict."""
    return {
        state: f'[{style}]\u25cf[/{style}]'
        for state, style in ci_styles(ts).items()
    }


def ci_check_styles(ts: Dict[str, str]) -> Dict[str, str]:
    """Return CI check detail styles from a resolved theme dict."""
    return {
        'pending': 'dim',
        'success': ts['success'],
        'warning': ts['warning'],
        'fail': f"bold {ts['error']}",
    }


def reviewer_colours(ts: Dict[str, str]) -> List[str]:
    """Return the reviewer colour palette from a resolved theme dict.

    Index 0 is always the current user; the rest cycle for others.
    """
    return [
        ts['warning'],      # index 0: current user (warm/distinct)
        ts['accent'],
        ts['secondary'],
        ts['error'],
        ts['primary'],
    ]


class _quiet_worker:
    """Context manager that silences b4 logger output from the current thread.

    Worker threads in Textual run while the TUI owns the terminal, so any
    logger output would overwrite the screen.  This installs a thread-aware
    filter on the b4 logger's handlers for the duration of the block.
    """

    def __enter__(self) -> '_quiet_worker':
        import logging
        import threading

        tid = threading.current_thread().ident

        class _Filter(logging.Filter):
            def filter(self, record: logging.LogRecord) -> bool:
                return record.thread != tid

        self._filt = _Filter()
        self._logger = logging.getLogger('b4')
        for h in self._logger.handlers:
            h.addFilter(self._filt)
        return self

    def __exit__(self, *exc: object) -> None:
        for h in self._logger.handlers:
            h.removeFilter(self._filt)


class CheckRunnerMixin:
    """Mixin providing CI check execution for Textual App subclasses.

    Subclasses must implement :meth:`_get_check_context` to supply the
    message-id, subject, and change-id for the series to check.  All UI
    interaction (loading overlay, results modal) is handled here.
    """

    _check_loading: Optional[Any] = None

    # -- interface for subclasses ------------------------------------------

    def _get_check_context(self) -> Optional[Tuple[str, str, str]]:
        """Return ``(message_id, series_subject, change_id)`` or *None*.

        Return *None* when no series is selected or checks cannot run.
        """
        raise NotImplementedError

    # -- public action -----------------------------------------------------

    def action_check(self) -> None:
        """Run CI checks on the current series."""
        self._run_checks(force=False)

    # -- helpers -----------------------------------------------------------

    def _run_checks(self, force: bool = False) -> None:
        """Show loading overlay and launch the check worker thread."""
        ctx = self._get_check_context()
        if ctx is None:
            return
        message_id, series_subject, change_id = ctx
        if not message_id:
            self.notify('No message-id for this series', severity='error')  # type: ignore[attr-defined]
            return
        from b4.review_tui._modals import CheckLoadingScreen
        self._check_loading = CheckLoadingScreen()
        self.push_screen(self._check_loading)  # type: ignore[attr-defined]
        self.run_worker(  # type: ignore[attr-defined]
            lambda: self._fetch_and_check(message_id, series_subject,
                                          change_id=change_id, force=force),
            name='_check_worker', thread=True)

    def _dismiss_loading(self, msg: str = '', severity: str = '') -> None:
        """Dismiss the check loading screen and optionally notify."""
        def _do() -> None:
            if self._check_loading is not None and self._check_loading.is_attached:
                self._check_loading.dismiss(None)
            if msg:
                self.notify(msg, severity=severity)  # type: ignore[attr-defined]
        self.app.call_from_thread(_do)  # type: ignore[attr-defined]

    def _update_loading(self, text: str) -> None:
        """Update the loading screen status text."""
        def _do() -> None:
            if self._check_loading is not None and self._check_loading.is_attached:
                self._check_loading.update_status(text)
        self.app.call_from_thread(_do)  # type: ignore[attr-defined]

    def _fetch_and_check(self, message_id: str, series_subject: str,
                         change_id: str = '', force: bool = False) -> None:
        """Fetch thread, run checks, and push results modal (worker thread)."""
        import b4.review.checks as checks
        from b4.review_tui._modals import TrackingCheckResultsScreen

        checks.clear_sashiko_cache()
        perpatch_cmds, series_cmds = checks.load_check_cmds()
        if not perpatch_cmds and not series_cmds:
            self._dismiss_loading('No check commands configured', 'warning')
            return

        topdir = b4.git_get_toplevel()
        if not topdir:
            self._dismiss_loading('Not in a git repository', 'error')
            return

        # Patchwork config for _builtin_patchwork
        config = b4.get_main_config()
        pwkey = str(config.get('pw-key', ''))
        pwurl = str(config.get('pw-url', ''))

        # Dump tracking data to a temp file for external scripts
        extra_env: Dict[str, str] = {}
        tracking_file: Optional[str] = None
        blob_sha = ''
        if change_id:
            review_branch = f'b4/review/{change_id}'
            try:
                _cover, tracking = b4.review.load_tracking(topdir, review_branch)
                blob_sha = tracking.get('series', {}).get('thread-blob', '')
                fd, tracking_file = tempfile.mkstemp(prefix='b4-tracking-', suffix='.json')
                with os.fdopen(fd, 'w') as fp:
                    json.dump(tracking, fp, indent=2)
                extra_env['B4_TRACKING_FILE'] = tracking_file
            except SystemExit:
                pass

        # Fetch the thread (local blob first, then lore)
        self._update_loading('Loading thread\u2026')
        with _quiet_worker():
            msgs = get_thread_msgs(topdir, message_id,
                                   blob_sha=blob_sha, quiet=True)
        if not msgs:
            self._dismiss_loading('Could not fetch thread from lore', 'error')
            return

        # Separate patches from non-patches using LoreSubject
        cover_msg: Optional[Tuple[str, email.message.EmailMessage]] = None
        patches: List[Tuple[str, email.message.EmailMessage]] = []
        for msg in msgs:
            subject = msg.get('subject', '')
            if not subject:
                continue
            lsubj = b4.LoreSubject(subject)
            msgid = msg.get('message-id', '').strip('<> ')
            if not msgid:
                continue
            if lsubj.counter == 0 and lsubj.expected > 0:
                cover_msg = (msgid, msg)
            elif lsubj.patch and not lsubj.reply:
                patches.append((msgid, msg))

        if not patches and not cover_msg:
            self._dismiss_loading('No patches found in thread', 'error')
            return

        # Sort patches by counter
        patches.sort(key=lambda p: b4.LoreSubject(p[1].get('subject', '')).counter)

        # Open or create cache DB
        conn = checks.get_db()
        checks.cleanup_old(conn)

        expected = patches[0][1].get('subject', '') if patches else ''
        lsubj0 = b4.LoreSubject(expected)
        num_patches = lsubj0.expected if lsubj0.expected > 0 else len(patches)

        # Build patch labels and subjects
        patch_labels: List[str] = []
        patch_subjects: List[str] = []
        ordered_msgs: List[Tuple[str, email.message.EmailMessage]] = []
        if cover_msg:
            patch_labels.append(f'0/{num_patches}')
            patch_subjects.append(b4.LoreSubject(cover_msg[1].get('subject', '')).subject)
            ordered_msgs.append(cover_msg)
        for idx, (mid, msg) in enumerate(patches, 1):
            patch_labels.append(f'{idx}/{num_patches}')
            patch_subjects.append(b4.LoreSubject(msg.get('subject', '')).subject)
            ordered_msgs.append((mid, msg))

        # Check cache first (clear and skip when force-rerunning)
        all_msgids = [mid for mid, _ in ordered_msgs]
        cached: Dict[str, List[Dict[str, str]]] = {}
        if force:
            checks.delete_results(conn, all_msgids)
        else:
            cached = checks.get_cached_results(conn, all_msgids)

        all_tools: set[str] = set()
        matrix: Dict[Tuple[int, str], Dict[str, str]] = {}

        # Populate matrix from cache
        for pidx, (mid, _msg) in enumerate(ordered_msgs):
            for result in cached.get(mid, []):
                tool = result['tool']
                all_tools.add(tool)
                matrix[(pidx, tool)] = result

        # Collect new results for batch DB storage
        new_results: Dict[str, List[Dict[str, str]]] = {}

        # Determine which patches still need checking
        if perpatch_cmds:
            start_idx = 1 if cover_msg else 0
            unchecked: List[Tuple[int, str, email.message.EmailMessage]] = []
            for pidx_offset, (mid, _msg) in enumerate(ordered_msgs[start_idx:]):
                pidx = pidx_offset + start_idx
                if mid not in cached:
                    unchecked.append((pidx, mid, _msg))

            for pidx, mid, _msg in unchecked:
                label = patch_labels[pidx]
                self._update_loading(f'Running checks\u2026 {label}')
                single_results = checks.run_perpatch_checks(
                    [(mid, _msg)], perpatch_cmds, topdir, pwkey, pwurl,
                    extra_env=extra_env)
                for result in single_results.get(mid, []):
                    tool = result['tool']
                    all_tools.add(tool)
                    matrix[(pidx, tool)] = result
                    new_results.setdefault(mid, []).append(result)

        # Run per-series checks (only if not cached)
        if series_cmds:
            target = cover_msg if cover_msg else (ordered_msgs[0] if ordered_msgs else None)
            if target and target[0] not in cached:
                self._update_loading('Running series checks\u2026')
                series_results = checks.run_series_checks(
                    target, series_cmds, topdir, pwkey, pwurl,
                    extra_env=extra_env)
                cover_idx = 0
                for result in series_results:
                    tool = result['tool']
                    all_tools.add(tool)
                    matrix[(cover_idx, tool)] = result
                    new_results.setdefault(target[0], []).append(result)

        # Batch-store all new results
        for mid, results in new_results.items():
            checks.store_results(conn, mid, results)
        conn.close()

        # Clean up the tracking data temp file
        if tracking_file:
            try:
                os.unlink(tracking_file)
            except OSError:
                pass

        # Build title and swap loading screen for results
        title = f'CI Check Results: {series_subject}'
        tools_sorted = sorted(all_tools)

        def _on_result(result: Optional[str]) -> None:
            if result == 'rerun':
                self._run_checks(force=True)

        def _push_modal() -> None:
            if self._check_loading is not None and self._check_loading.is_attached:
                self._check_loading.dismiss(None)
            self.push_screen(TrackingCheckResultsScreen(  # type: ignore[attr-defined]
                title, patch_labels, patch_subjects, tools_sorted, matrix),
                callback=_on_result)

        self.app.call_from_thread(_push_modal)  # type: ignore[attr-defined]


def _wait_for_enter() -> None:
    try:
        input('Press Enter to continue...')
    except (KeyboardInterrupt, EOFError):
        pass


def _make_initials(name: str) -> str:
    """Derive initials from a maintainer name.

    - "Konstantin Ryabitsev" -> "KR"
    - "Foo Bar Barski"       -> "FBB"
    - "FooBarski"            -> "FB"  (CamelCase single word)
    - ""                     -> "?"
    """
    name = name.strip()
    if not name:
        return 'me'
    parts = name.split()
    if len(parts) > 1:
        return ''.join(p[0].upper() for p in parts if p)
    # Single word: extract uppercase letters for CamelCase
    uppers = [c for c in parts[0] if c.isupper()]
    if len(uppers) >= 2:
        return ''.join(uppers)
    # Fallback: first two characters
    return parts[0][:2].upper()


def _has_review_data(reviews: Dict[str, Dict[str, Any]]) -> bool:
    """Return True if any reviewer has trailers, reply, comments, or a note."""
    return any(
        r.get('trailers') or r.get('reply', '') or r.get('comments') or r.get('note', '')
        for r in reviews.values()
    )


def _strip_attribution(body: str) -> str:
    """Strip a leading attribution line if present.

    An attribution line is the first non-blank line (possibly wrapped
    across consecutive non-blank lines) that ends with ':' and is
    followed (after optional blank lines) by a quoted line ('> ').
    """
    lines = body.splitlines()
    # Find the first non-blank line
    first_idx = None
    for i, ln in enumerate(lines):
        if ln.strip():
            first_idx = i
            break
    if first_idx is None:
        return body
    # Walk consecutive non-blank lines looking for one ending with ':'
    attr_end = None
    for i in range(first_idx, len(lines)):
        if not lines[i].strip():
            break
        if lines[i].strip().endswith(':'):
            attr_end = i
            break
    if attr_end is None:
        return body
    # Check that the next non-blank line starts with '>'
    for ln in lines[attr_end + 1:]:
        if ln.strip():
            if ln.startswith('> ') or ln.strip() == '>':
                remaining = lines[attr_end + 1:]
                while remaining and (not remaining[0].strip() or remaining[0].strip() == '>'):
                    remaining.pop(0)
                return '\n'.join(remaining)
            break
    return body


def _write_followup_comments(
    viewer: 'RichLog',
    fc_list: List[Dict[str, Any]],
    comment_positions: List[int],
    header_position_map: Optional[Dict[int, Dict[str, Any]]] = None,
    ts: Optional[Dict[str, str]] = None,
) -> None:
    """Render follow-up comments at the bottom of the viewer.

    Each entry has 'body', 'fromname', 'fromemail', 'date', and optionally
    'depth' (0 = direct reply to patch, N = N hops through follow-up replies).
    Entries with depth > 0 are indented to visually show threading.
    If *header_position_map* is provided, it is populated with
    {viewer_line: entry} for each panel header row, enabling click-to-reply.
    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    """
    if not fc_list:
        return
    rev_palette = reviewer_colours(ts) if ts else [
        'dark_goldenrod', 'dark_cyan',
        'dark_magenta', 'dark_red', 'dark_blue',
    ]
    fc_emails = sorted({e['fromemail'] for e in fc_list})
    colour_map: Dict[str, str] = {}
    for ci, em in enumerate(fc_emails):
        colour_map[em] = rev_palette[1 + (ci % (len(rev_palette) - 1))]
    viewer.write(Text(''))
    viewer.write(Rule(title='follow-ups', style='dim'))
    viewer.write(Text(''))
    for e in sorted(fc_list, key=lambda x: x['date']):
        fromname = e['fromname']
        colour = colour_map.get(e['fromemail'], rev_palette[1])
        line_pos = len(viewer.lines)
        comment_positions.append(line_pos)
        if header_position_map is not None:
            header_position_map[line_pos] = e
        body = _strip_attribution(e['body'])
        body_text = Text()
        body_text.append(f'From:  {fromname} <{e["fromemail"]}>\n', style='bold')
        body_text.append(f'Date:  {e["date"].strftime("%Y-%m-%d %H:%M %z")}\n', style='bold')
        if msgid := e.get('msgid', ''):
            body_text.append(f'Msgid: <{msgid}>\n', style='bold')
        body_text.append('\n')
        for line in body.splitlines():
            if line.startswith('>'):
                body_text.append(line, style='dim')
            else:
                body_text.append(line)
            body_text.append('\n')
        depth = e.get('depth', 0)
        title_text = Text()
        title_text.append(fromname)
        title_text.append('  ↩', style='dim')
        panel = Panel(
            body_text,
            box=box.ROUNDED,
            border_style=colour,
            title=title_text,
            title_align='left',
            expand=True,
            padding=(0, 1),
        )
        if depth > 0:
            viewer.write(Padding(panel, pad=(0, 0, 0, depth * 2)))
        else:
            viewer.write(panel)


_FOLLOWUP_MAX_DEPTH = 5


def _get_followup_depth(
    in_reply_to: Optional[str],
    patch_msgids: Dict[str, int],
    msgid_map: Dict[str, 'b4.LoreMessage'],
) -> int:
    """Return the threading depth of a follow-up relative to its patch.

    Depth 0 = direct reply to a patch; depth N = N hops through follow-up
    replies. Capped at _FOLLOWUP_MAX_DEPTH to prevent runaway indentation.
    """
    depth = 0
    seen: Set[str] = set()
    current = in_reply_to
    while current and current not in seen:
        if current in patch_msgids:
            break
        seen.add(current)
        lmsg = msgid_map.get(current)
        if lmsg is None:
            break
        depth += 1
        current = lmsg.in_reply_to
    return min(depth, _FOLLOWUP_MAX_DEPTH)


def _resolve_patch_for_followup(
    in_reply_to: Optional[str],
    patch_msgids: Dict[str, int],
    msgid_map: Dict[str, 'b4.LoreMessage'],
) -> Optional[int]:
    """Walk the in_reply_to chain to find which patch a follow-up belongs to.

    Returns the display index (0=cover, 1..N=patches) or None.
    """
    seen: Set[str] = set()
    current = in_reply_to
    while current and current not in seen:
        if current in patch_msgids:
            return patch_msgids[current]
        seen.add(current)
        lmsg = msgid_map.get(current)
        if lmsg is None:
            break
        current = lmsg.in_reply_to
    return None


def _chain_has_additional_patch(
    in_reply_to: Optional[str],
    patch_msgids: Dict[str, int],
    msgid_map: Dict[str, 'b4.LoreMessage'],
) -> bool:
    """Check whether the in-reply-to chain passes through a message that
    contains its own diff before reaching a known series patch.

    Such a message is an additional patch posted as a follow-up.
    Replies to it discuss that new code, so their quoted diffs should not
    be treated as inline reviews of the original series patch.
    """
    seen: Set[str] = set()
    current = in_reply_to
    while current and current not in seen:
        if current in patch_msgids:
            return False
        seen.add(current)
        lmsg = msgid_map.get(current)
        if lmsg is None:
            break
        if lmsg.has_diff:
            return True
        current = lmsg.in_reply_to
    return False


def _write_comments(
    viewer: 'RichLog',
    entries: List[Tuple[str, str, str]],
    ts: Optional[Dict[str, str]] = None,
) -> None:
    """Write review comment entries to *viewer* as bordered panels.

    Each entry is a (name, colour, text) tuple.  Comments from
    the same diff line are rendered as separate panels.
    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    """
    bg = f"on {ts['panel']}" if ts else 'on grey11'
    for name, colour, text in entries:
        panel = Panel(
            Text(text),
            box=box.ROUNDED,
            border_style=colour,
            subtitle=name,
            subtitle_align='right',
            expand=False,
            padding=(1, 1),
            style=bg,
        )
        viewer.write(panel)


def _write_followup_trailers(
    viewer: 'RichLog',
    followups: List[Dict[str, Any]],
    existing: Optional[Set[str]] = None,
    ts: Optional[Dict[str, str]] = None,
) -> None:
    """Write follow-up trailers as a panel in the diff viewer.

    Each followup dict has 'link', 'fromname', 'fromemail', and 'trailers'.
    Trailers already present in *existing* (lowercased) are skipped.
    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    """
    if existing is None:
        existing = set()
    trailer_color = ts['success'] if ts else 'green'
    lines = Text()
    for followup in followups:
        link = followup.get('link', '')
        for tstr in followup.get('trailers', []):
            if tstr.lower() in existing:
                continue
            existing.add(tstr.lower())
            if lines.plain:
                lines.append('\n')
            lines.append(tstr, style=f'bold {trailer_color}')
            if link:
                lines.append(f'\n  {link}', style=f'dim {trailer_color} link {link}')
    if not lines.plain:
        return
    viewer.write(Text(''))
    panel = Panel(
        lines,
        box=box.ROUNDED,
        title='+++',
        title_align='left',
        border_style=trailer_color,
        expand=False,
        padding=(0, 1),
    )
    viewer.write(panel)


def _write_diff_line(
    viewer: 'RichLog', line: str,
    ts: Optional[Dict[str, str]] = None,
) -> None:
    """Write a single diff line to a RichLog with appropriate colouring.

    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    """
    if line.startswith(('diff --git ', '--- ', '+++ ')):
        viewer.write(Text(line, style='bold'))
    elif line.startswith('@@'):
        viewer.write(Text(line, style=f"bold {ts['accent']}" if ts else 'bold cyan'))
    elif line.startswith('+'):
        viewer.write(Text(line, style=ts['success'] if ts else 'green'))
    elif line.startswith('-'):
        viewer.write(Text(line, style=ts['error'] if ts else 'red'))
    else:
        viewer.write(Text(line))


def _render_email_to_viewer(
    viewer: 'RichLog', msg: email.message.EmailMessage,
    ts: Optional[Dict[str, str]] = None,
) -> None:
    """Render an EmailMessage into a RichLog, headers first then body.

    Subject, To and Cc labels are bold.  Address headers are word-wrapped
    using LoreMessage.wrap_header.  Quoted body lines (>) are dim cyan,
    separator lines (---) are dim.
    *ts* is a resolved theme styles dict from :func:`resolve_styles`.
    """
    for hdr in ('From', 'Subject', 'To', 'Cc'):
        val = msg[hdr]
        if not val:
            continue
        val = str(val)
        if hdr.lower() in ('to', 'cc'):
            wrapped = b4.LoreMessage.wrap_header(
                (hdr, val), transform='decode').decode(errors='replace')
            first_line, *rest = wrapped.splitlines()
            colon = first_line.find(':')
            hdr_text = Text()
            if colon >= 0:
                hdr_text.append(first_line[:colon + 1], style='bold')
                hdr_text.append(first_line[colon + 1:])
            else:
                hdr_text.append(first_line)
            for r in rest:
                hdr_text.append('\n')
                hdr_text.append(r)
            viewer.write(hdr_text)
        else:
            hdr_text = Text()
            hdr_text.append(f'{hdr}:', style='bold')
            hdr_text.append(f' {val}')
            viewer.write(hdr_text)
    viewer.write('')
    payload = msg.get_payload(decode=True)
    body = payload.decode(errors='replace') if isinstance(payload, bytes) else str(payload or '')
    for line in body.splitlines():
        if line.startswith('>'):
            viewer.write(Text(line, style=f"dim {ts['accent']}" if ts else 'dim cyan'))
        elif line.startswith('---'):
            viewer.write(Text(line, style='dim'))
        else:
            viewer.write(Text(line))


def _suspend_to_shell(hint: str = 'b4 review', cwd: Optional[str] = None) -> None:
    """Spawn an interactive sub-shell with a PS1 hint.

    For bash and zsh, a temporary rc file is used so the user's normal
    configuration is loaded first and then the prompt is prefixed with
    a short marker.  For other shells the B4_REVIEW environment variable
    is set so the user can incorporate it into their own prompt.
    """
    logger.info('---')
    logger.info('You are now in shell mode. You can execute git commands or run checks.')
    logger.info('Cosmetic commit edits (reword subjects, fix trailers) are fine;')
    logger.info('b4 will reconcile tracking data when you return.')
    logger.info('Do NOT add, remove, squash, or reorder commits.')
    logger.info('When done, Ctrl-d to return to review UI.')
    logger.info('---')

    shell = os.environ.get('SHELL', '/bin/sh')
    shellname = os.path.basename(shell)
    env = os.environ.copy()
    env['B4_REVIEW'] = hint

    if shellname == 'bash':
        bashrc = os.path.expanduser('~/.bashrc')
        source = f'[ -f {bashrc} ] && . {bashrc}\n'
        source += f'PS1="({hint}) $PS1"\n'
        with tempfile.NamedTemporaryFile(mode='w', prefix='b4-shell-',
                                         suffix='.sh', delete=False) as rcf:
            rcf.write(source)
            rcfile = rcf.name
        try:
            subprocess.run([shell, '--rcfile', rcfile], env=env, cwd=cwd)
        finally:
            os.unlink(rcfile)
    elif shellname == 'zsh':
        real_zdotdir = os.environ.get('ZDOTDIR', os.path.expanduser('~'))
        with tempfile.TemporaryDirectory(prefix='b4-shell-') as tmpdir:
            zshrc = os.path.join(tmpdir, '.zshrc')
            with open(zshrc, 'w') as f:
                f.write(f'ZDOTDIR="{real_zdotdir}"\n')
                f.write('[ -f "$ZDOTDIR/.zshrc" ] && . "$ZDOTDIR/.zshrc"\n')
                f.write(f'PS1="({hint}) $PS1"\n')
            env['ZDOTDIR'] = tmpdir
            subprocess.run([shell], env=env, cwd=cwd)
    else:
        subprocess.run([shell], env=env, cwd=cwd)


def _addrs_to_lines(header_str: str) -> str:
    """Parse a comma-separated address header into one-per-line display."""
    if not header_str:
        return ''
    pairs = email.utils.getaddresses([header_str])
    lines = []
    for name, addr in pairs:
        if not addr:
            continue
        if name and name != addr:
            lines.append(f'{name} <{addr}>')
        else:
            lines.append(addr)
    return '\n'.join(lines)


def _lines_to_header(text: str) -> str:
    """Parse one-per-line addresses back to a comma-separated header string."""
    text = text.strip()
    if not text:
        return ''
    # getaddresses expects a list of header strings; join lines with commas
    pairs = email.utils.getaddresses(text.splitlines())
    pairs = [(n, a) for n, a in pairs if a]
    return b4.format_addrs(pairs, clean=False)


def _validate_addrs(text: str) -> Optional[str]:
    """Return an error message if any line has an invalid address, or None."""
    text = text.strip()
    if not text:
        return None
    for line in text.splitlines():
        line = line.strip()
        if not line:
            continue
        pairs = email.utils.getaddresses([line])
        if not pairs:
            return f'Cannot parse: {line}'
        for _name, addr in pairs:
            if not addr or '@' not in addr:
                return f'Invalid address: {line}'
    return None


def gather_attestation_info(lser: b4.LoreSeries) -> Dict[str, Any]:
    """Gather attestation and applicability information for a series.

    Args:
        lser: The LoreSeries to check

    Returns:
        Dict with keys:
            - total: Total number of patches
            - passing: Number of passing patches
            - critical: If True, hardfail policy triggered
            - same_attestation: If True, all patches have same attestations
            - attestations: Common attestations (if same_attestation) or None
            - per_patch: List of per-patch info (if not same_attestation)
            - base_commit: The base commit (or None)
            - base_exists: Whether the base commit exists in the repo
            - applies_clean: Whether the series applies cleanly (None if not checked)
            - apply_checked: Number of blobs checked
            - apply_mismatches: Number of mismatched blobs
    """
    config = b4.get_main_config()
    attpolicy = str(config.get('attestation-policy', 'softfail'))

    # Check base commit and applicability
    topdir = b4.git_get_toplevel()
    base_commit = lser.base_commit
    base_exists = False
    applies_clean = None
    apply_checked = 0
    apply_mismatches = 0

    if topdir:
        # Ensure indexes are populated for applicability check
        if lser.indexes is None:
            lser.populate_indexes()

        if base_commit:
            base_exists = b4.git_commit_exists(topdir, base_commit)

        # Check applicability
        if lser.indexes:
            check_at = None
            if base_commit and base_exists:
                check_at = base_commit
            else:
                # No base commit or base doesn't exist - check against HEAD
                check_at = 'HEAD'

            try:
                apply_checked, mismatches = lser.check_applies_clean(topdir, at=check_at)
                apply_mismatches = len(mismatches)
                applies_clean = (apply_mismatches == 0)
            except Exception:
                pass

    if attpolicy == 'off':
        return {
            'total': 0,
            'passing': 0,
            'critical': False,
            'same_attestation': True,
            'attestations': [],
            'per_patch': [],
            'base_commit': base_commit,
            'base_exists': base_exists,
            'applies_clean': applies_clean,
            'apply_checked': apply_checked,
            'apply_mismatches': apply_mismatches,
        }

    try:
        maxdays = int(str(config.get('attestation-staleness-days', '0')))
    except ValueError:
        maxdays = 0

    per_patch = []
    any_critical = False
    total_passing = 0
    ref_attestations: Optional[List[Dict[str, Any]]] = None
    same_attestation = True

    total = lser.expected
    width = len(str(total))
    for idx, lmsg in enumerate(lser.patches[1:], start=1):
        patch_idx = f'{idx:0{width}d}/{total:0{width}d}'

        if lmsg is None:
            per_patch.append({
                'index': patch_idx,
                'passing': False,
                'attestations': [{'status': 'missing', 'identity': 'Patch not available', 'passing': False}],
            })
            same_attestation = False
            continue

        attestations, overall_passing, critical = lmsg.get_attestation_status(attpolicy, maxdays)
        if critical:
            any_critical = True

        if overall_passing:
            total_passing += 1

        # Check if attestation is the same as previous patches (compare by identity set)
        if ref_attestations is None:
            ref_attestations = attestations
        else:
            ref_ids = {a['identity'] for a in ref_attestations}
            cur_ids = {a['identity'] for a in attestations}
            if ref_ids != cur_ids:
                same_attestation = False

        per_patch.append({
            'index': patch_idx,
            'passing': overall_passing,
            'attestations': attestations,
        })

    return {
        'total': len(per_patch),
        'passing': total_passing,
        'critical': any_critical,
        'same_attestation': same_attestation,
        'attestations': ref_attestations or [],
        'per_patch': per_patch,
        'base_commit': base_commit,
        'base_exists': base_exists,
        'applies_clean': applies_clean,
        'apply_checked': apply_checked,
        'apply_mismatches': apply_mismatches,
    }


class JKListNavMixin:
    """Mixin providing j/k cursor navigation for a named ListView.

    Classes using this mixin must set ``_list_id`` to the DOM id of the
    target :class:`ListView` (e.g. ``'#action-list'``).
    """

    _list_id: str = ''

    def action_cursor_down(self) -> None:
        lv = self.query_one(self._list_id, ListView)  # type: ignore[attr-defined]
        if lv.index is not None and lv.index < len(lv.children) - 1:
            lv.index += 1

    def action_cursor_up(self) -> None:
        lv = self.query_one(self._list_id, ListView)  # type: ignore[attr-defined]
        if lv.index is not None and lv.index > 0:
            lv.index -= 1


_SENTINEL = object()


class SeparatedFooter(Footer):
    """Footer that shows full descriptions and a vertical separator between groups."""

    DEFAULT_CSS = """
    SeparatedFooter FooterKey.-group-first {
        border-left: vkey $foreground 20%;
    }
    SeparatedFooter:ansi {
        background: ansi_bright_black;
        .footer-key--key {
            background: ansi_bright_black;
        }
        .footer-key--description {
            background: ansi_bright_black;
        }
    }
    SeparatedFooter:ansi FooterKey.-group-first {
        border-left: vkey ansi_default;
    }
    """

    def compose(self) -> ComposeResult:
        if not self._bindings_ready:
            return
        active_bindings = self.screen.active_bindings
        bindings = [
            (binding, enabled, tooltip)
            for (_, binding, enabled, tooltip) in active_bindings.values()
            if binding.show
        ]
        action_to_bindings: defaultdict[str, list[tuple[Binding, bool, str]]]
        action_to_bindings = defaultdict(list)
        for binding, enabled, tooltip in bindings:
            action_to_bindings[binding.action].append((binding, enabled, tooltip))

        self.styles.grid_size_columns = len(action_to_bindings)

        group_map = getattr(self.app, 'BINDING_GROUPS', {})
        prev_group: object = _SENTINEL
        for multi_bindings_list in action_to_bindings.values():
            binding, enabled, tooltip = multi_bindings_list[0]
            cur_group = group_map.get(binding.action)
            is_first = prev_group is not _SENTINEL and cur_group != prev_group
            prev_group = cur_group
            classes = '-group-first' if is_first else ''
            yield FooterKey(
                binding.key,
                self.app.get_key_display(binding),
                binding.description,
                binding.action,
                disabled=not enabled,
                tooltip=tooltip,
                classes=classes,
            ).data_bind(compact=Footer.compact)

        if self.show_command_palette and self.app.ENABLE_COMMAND_PALETTE:
            try:
                _node, binding, enabled, tooltip = active_bindings[
                    self.app.COMMAND_PALETTE_BINDING
                ]
            except KeyError:
                pass
            else:
                yield FooterKey(
                    binding.key,
                    self.app.get_key_display(binding),
                    binding.description,
                    binding.action,
                    classes='-command-palette',
                    disabled=not enabled,
                    tooltip=binding.tooltip or binding.description,
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024 by the Linux Foundation
#
"""Tests for ``b4 review show-info``."""
import json
import pytest


import b4
import b4.review
from b4.review._review import (
    get_review_info,
    show_review_info,
    list_review_branches,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _create_review_branch(gitdir: str, change_id: str,
                          identifier: str = 'test-project',
                          revision: int = 1,
                          status: str = 'reviewing',
                          subject: str = 'Test series',
                          sender_name: str = 'Test Author',
                          sender_email: str = 'test@example.com',
                          link: str = '',
                          num_real_commits: int = 0) -> str:
    """Create a fake b4 review branch with a proper tracking commit.

    When *num_real_commits* > 0, that many empty commits are created between
    the base and the tracking commit so ``commit-{hash}`` keys appear.

    Returns the branch name.
    """
    branch_name = f'b4/review/{change_id}'
    # Get current HEAD as base
    ecode, base_sha = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
    assert ecode == 0
    base_sha = base_sha.strip()

    # Create the branch at HEAD
    ecode, _ = b4.git_run_command(gitdir, ['branch', branch_name, base_sha])
    assert ecode == 0

    # Check out the branch to add commits
    ecode, _ = b4.git_run_command(gitdir, ['checkout', branch_name])
    assert ecode == 0

    # Create real patch commits if requested
    first_patch_commit = None
    for i in range(num_real_commits):
        ecode, _ = b4.git_run_command(
            gitdir, ['commit', '--allow-empty', '-m', f'patch {i+1}: do thing {i+1}'])
        assert ecode == 0
        if i == 0:
            ecode, sha = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
            assert ecode == 0
            first_patch_commit = sha.strip()

    if first_patch_commit is None:
        first_patch_commit = base_sha

    # Build tracking metadata
    trk = {
        'series': {
            'identifier': identifier,
            'change-id': change_id,
            'revision': revision,
            'status': status,
            'subject': subject,
            'fromname': sender_name,
            'fromemail': sender_email,
            'expected': max(num_real_commits, 1),
            'complete': True,
            'base-commit': base_sha,
            'prerequisite-commits': [],
            'first-patch-commit': first_patch_commit,
            'link': link,
            'header-info': {},
        },
        'followups': [],
        'patches': [],
    }
    commit_msg = f'{subject}\n\n{b4.review.make_review_magic_json(trk)}'

    # Create the tracking commit
    ecode, _ = b4.git_run_command(
        gitdir, ['commit', '--allow-empty', '-m', commit_msg])
    assert ecode == 0

    # Go back to master
    ecode, _ = b4.git_run_command(gitdir, ['checkout', 'master'])
    assert ecode == 0

    return branch_name


# ---------------------------------------------------------------------------
# TestGetReviewInfo
# ---------------------------------------------------------------------------

class TestGetReviewInfo:

    def test_basic_info(self, gitdir: str) -> None:
        branch = _create_review_branch(gitdir, 'basic-change-id',
                                       subject='Basic test series',
                                       status='reviewing')
        info = get_review_info(gitdir, branch)

        assert info['branch'] == branch
        assert info['change-id'] == 'basic-change-id'
        assert info['status'] == 'reviewing'
        assert info['subject'] == 'Basic test series'
        assert info['revision'] == 1
        assert info['complete'] is True
        assert info['num-prereqs'] == 0
        assert info['base-commit'] is not None
        assert info['first-patch-commit'] is not None

    def test_sender_format(self, gitdir: str) -> None:
        branch = _create_review_branch(gitdir, 'sender-test',
                                       sender_name='Alice Author',
                                       sender_email='alice@example.com')
        info = get_review_info(gitdir, branch)
        assert info['sender'] == 'Alice Author <alice@example.com>'

    def test_commit_keys(self, gitdir: str) -> None:
        branch = _create_review_branch(gitdir, 'commit-keys-test',
                                       num_real_commits=3)
        info = get_review_info(gitdir, branch)

        assert info['num-patches'] == 3
        commit_keys = [k for k in info if k.startswith('commit-')]
        assert len(commit_keys) == 3
        # Each commit key should have a subject value
        for k in commit_keys:
            val = info[k]
            assert isinstance(val, str)
            assert len(val) > 0


# ---------------------------------------------------------------------------
# TestShowReviewInfo
# ---------------------------------------------------------------------------

class TestShowReviewInfo:

    def test_all_keys(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'show-all-test', subject='All keys test')
        show_review_info('b4/review/show-all-test:_all')
        out = capsys.readouterr().out
        assert 'branch: b4/review/show-all-test' in out
        assert 'change-id: show-all-test' in out
        assert 'subject: All keys test' in out

    def test_single_key(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'single-key-test', status='applied')
        show_review_info('b4/review/single-key-test:status')
        out = capsys.readouterr().out
        assert out.strip() == 'applied'

    def test_named_branch(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        branch = _create_review_branch(gitdir, 'named-branch-test')
        show_review_info(branch)
        out = capsys.readouterr().out
        assert 'branch: b4/review/named-branch-test' in out

    def test_shorthand(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'shorthand-test')
        show_review_info('shorthand-test:status')
        out = capsys.readouterr().out
        assert out.strip() == 'reviewing'

    def test_branch_not_found(self, gitdir: str) -> None:
        with pytest.raises(SystemExit):
            show_review_info('nonexistent-branch:status')

    def test_not_review_branch(self, gitdir: str) -> None:
        with pytest.raises(SystemExit):
            show_review_info('master:status')

    def test_json_output(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'json-test', subject='JSON output test')
        show_review_info('b4/review/json-test:_all', as_json=True)
        out = capsys.readouterr().out
        data = json.loads(out)
        assert data['change-id'] == 'json-test'
        assert data['subject'] == 'JSON output test'


# ---------------------------------------------------------------------------
# TestListReviewBranches
# ---------------------------------------------------------------------------

class TestListReviewBranches:

    def test_list_multiple(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'list-alpha', subject='Alpha series')
        _create_review_branch(gitdir, 'list-bravo', subject='Bravo series')
        list_review_branches()
        out = capsys.readouterr().out
        assert 'list-alpha' in out
        assert 'list-bravo' in out

    def test_list_empty(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        list_review_branches()
        # Should not crash; logger.info prints the message
        out = capsys.readouterr().out
        assert out == ''  # message goes to logger, not stdout

    def test_list_json(self, gitdir: str, capsys: pytest.CaptureFixture[str]) -> None:
        _create_review_branch(gitdir, 'json-alpha', subject='Alpha JSON')
        _create_review_branch(gitdir, 'json-bravo', subject='Bravo JSON')
        list_review_branches(as_json=True)
        out = capsys.readouterr().out
        data = json.loads(out)
        assert isinstance(data, list)
        assert len(data) == 2
        change_ids = {d['change-id'] for d in data}
        assert 'json-alpha' in change_ids
        assert 'json-bravo' in change_ids


# ---------------------------------------------------------------------------
# TestTargetBranchInInfo
# ---------------------------------------------------------------------------

class TestTargetBranchInInfo:

    def test_target_branch_in_info(self, gitdir: str) -> None:
        """Branch with target-branch in tracking data includes it in info."""
        branch_name = 'b4/review/target-info-test'
        # Get current HEAD as base
        ecode, base_sha = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
        assert ecode == 0
        base_sha = base_sha.strip()

        ecode, _ = b4.git_run_command(gitdir, ['branch', branch_name, base_sha])
        assert ecode == 0

        # Build tracking with target-branch set
        trk = {
            'series': {
                'identifier': 'test-project',
                'change-id': 'target-info-test',
                'revision': 1,
                'status': 'reviewing',
                'subject': 'Target info test',
                'fromname': 'Test',
                'fromemail': 'test@example.com',
                'expected': 1,
                'complete': True,
                'base-commit': base_sha,
                'prerequisite-commits': [],
                'first-patch-commit': base_sha,
                'target-branch': 'sound/for-next',
                'header-info': {},
            },
            'followups': [],
            'patches': [],
        }
        commit_msg = f'Target info test\n\n{b4.review.make_review_magic_json(trk)}'
        ecode, tree = b4.git_run_command(gitdir, ['rev-parse', f'{branch_name}^{{tree}}'])
        assert ecode == 0
        ecode, new_sha = b4.git_run_command(
            gitdir, ['commit-tree', tree.strip(), '-p', base_sha],
            stdin=commit_msg.encode())
        assert ecode == 0
        ecode, _ = b4.git_run_command(
            gitdir, ['update-ref', f'refs/heads/{branch_name}', new_sha.strip()])
        assert ecode == 0

        info = get_review_info(gitdir, branch_name)
        assert info['target-branch'] == 'sound/for-next'

    def test_target_branch_fallback(self, gitdir: str) -> None:
        """No per-series target + single config value = fallback shown."""
        from unittest.mock import patch as mock_patch
        branch = _create_review_branch(gitdir, 'target-fallback-test',
                                       subject='Fallback test')
        with mock_patch('b4.review.tracking.get_review_target_branch_default',
                        return_value='regulator/for-next'):
            info = get_review_info(gitdir, branch)
        assert info['target-branch'] == 'regulator/for-next'

    def test_target_branch_none(self, gitdir: str) -> None:
        """No per-series target + no config = None."""
        branch = _create_review_branch(gitdir, 'target-none-test',
                                       subject='None test')
        info = get_review_info(gitdir, branch)
        assert info['target-branch'] is None
